from typing import List, Dict, Optional, Union
from pydantic import BaseModel, Field, HttpUrl
from .job_task_types import WebDataSelector
from .enums import JobNavigationType

class LinearNavigationOptions(BaseModel):
    regex_pattern: Optional[str] = None
    next_page_selector: Optional[WebDataSelector] = None

class BreadthFirstNavigationOptions(BaseModel):
    regex_pattern: Optional[str] = None
    max_links_per_page: Optional[int] = Field(10, ge=1, le=100)
    skip_parent: Optional[bool] = None

class CustomPathNavigationOptions(BaseModel):
    urls: str

class JobNavigation(BaseModel):
    type: JobNavigationType
    options: Optional[Union[LinearNavigationOptions, BreadthFirstNavigationOptions, CustomPathNavigationOptions]] = None
    max_pages: Optional[int] = Field(50, ge=1, le=1000)

    class Config:
        json_encoders = {JobNavigationType: lambda v: v.value}



from typing import Optional, Union
from pydantic import BaseModel
from datetime import datetime
from .enums import JobSchedulePolicy

class OneTimeSchedule(BaseModel):
    timezone: str
    timestamp: datetime  # Automatically converts from string to datetime

class JobSchedule(BaseModel):
    policy: JobSchedulePolicy
    specification: Optional[
        Union[OneTimeSchedule, str]
    ] = None
    last_run: Optional[datetime] = None

    class Config:
        json_encoders = {JobSchedulePolicy: lambda v: v.value}





from typing import List, Optional, Union, Any
from pydantic import BaseModel
from .enums import (
    JobTaskType,
    WebDataSelectorType, SpecificDataExtractionType,
    SpecificDataExtractionOutputFormat, WebScreenCaptureFormat, 
    StructuredDataExtractionAttribute, PdfPaperFormat, PdfOrientation
)

class WebDataSelector(BaseModel):
    type: WebDataSelectorType
    value: str

    class Config:
        json_encoders = {WebDataSelectorType: lambda v: v.value}

class StructuredDataExtractionParameter(BaseModel):
    name: str
    selector: Optional[WebDataSelector] = None
    parent_selector: Optional[WebDataSelector] = None
    attribute: Optional[StructuredDataExtractionAttribute] = None
    custom_attribute: Optional[str] = None
    sample: Optional[str] = None

    class Config:
        json_encoders = {StructuredDataExtractionAttribute: lambda v: v.value}

class StructuredDataExtractionOptions(BaseModel):
    parameters: List[StructuredDataExtractionParameter]
    auto_populate_parameters: bool = True
    parse_with_ai: bool = False
    execute_with_ai: bool = False
    parent_selector: WebDataSelector = WebDataSelector( type="css", value="body" )

class FormDataExtractionOptions(BaseModel):
    selector: Optional[WebDataSelector] = None

class TableDataExtractionOptions(BaseModel):
    selector: Optional[WebDataSelector] = None

class PresetDataExtractionOptions(BaseModel):
    selector: Optional[WebDataSelector] = None

class CustomDataExtractionOptions(BaseModel):
    regex_pattern: Optional[str] = None
    selector: Optional[WebDataSelector] = None


class SpecificDataExtractionOptions(BaseModel):
    type: SpecificDataExtractionType
    options: Union[
        PresetDataExtractionOptions, 
        TableDataExtractionOptions, 
        FormDataExtractionOptions, 
        StructuredDataExtractionOptions,
        CustomDataExtractionOptions,
        None
    ]
    max_results: int = 100
    output_format: SpecificDataExtractionOutputFormat

    class Config:
        json_encoders = {
            SpecificDataExtractionType: lambda v: v.value,
            SpecificDataExtractionOutputFormat: lambda v: v.value
        }


class WebScreenCaptureOptions(BaseModel):
    format: Optional[WebScreenCaptureFormat] = WebScreenCaptureFormat.WEBP.value
    full_page: Optional[bool] = False
    omit_background: Optional[bool] = True
    pdf_format: Optional[PdfPaperFormat] = PdfPaperFormat.A4.value
    pdf_orientation: Optional[PdfOrientation] = PdfOrientation.LANDSCAPE.value
    pdf_print_background: Optional[bool] = True

    class Config:
        json_encoders = {
            PdfOrientation: lambda v: v.value,
            PdfPaperFormat: lambda v: v.value,
            WebScreenCaptureFormat: lambda v: v.value,
        }


class WebpageAsMarkdownOptions(BaseModel):
    only_main_content: Optional[bool] = True
    apply_smart_cleaner: Optional[bool] = True


class JobTaskBase(BaseModel):
    id: Optional[str] = None
    type: JobTaskType


class JobTask(JobTaskBase):
    options: Union[
        SpecificDataExtractionOptions, 
        WebpageAsMarkdownOptions, 
        WebScreenCaptureOptions,
        None
    ]

    class Config:
        json_encoders = {JobTaskType: lambda v: v.value}


class PageOutput(BaseModel):
    source: str
    content: str


class JobTaskOutput(BaseModel):
    task: JobTaskBase
    data: List[Union[PageOutput]]
    data_url: Optional[str] = None


from typing import List, Dict, Union, Any, Optional
from pydantic import BaseModel
from .job_types import Job, JobDetail, JobResult, JobStats
from .enums import (
    JobExecutionStatus, 
)

class BaseResponse(BaseModel):
    message: str

class PaginationResponse(BaseModel):
    page: int
    limit: int 
    total_count: int

class JobStatsResponse(BaseResponse):
    job_stats: JobStats

class JobEstimatedCostResponse(BaseResponse):
    job_estimated_cost: int

class JobCreationResponse(BaseResponse):
    job: Job

class JobExecutionResponse(BaseResponse):
    status: JobExecutionStatus
    job_result: Union[JobResult, Dict[str, Any]]

    class Config:
        json_encoders = {
            JobExecutionStatus: lambda v: v.value
        }

class GetJobResponse(BaseResponse):
    job: JobDetail

class GetJobsResponse(BaseResponse):
    jobs: List[Job]
    pagination: Optional[PaginationResponse] = None

class GetJobResultResponse(BaseResponse):
    job_result: JobResult



from typing import List, Dict, Optional
from pydantic import BaseModel
from datetime import datetime
from .enums import (
    JobStatus, PuppetDeviceType, PuppetRegion
)
from .job_task_types import JobTask, JobTaskOutput
from .job_schedule_types import JobSchedule
from .job_navigation_types import JobNavigation


class WebAgentConfig(BaseModel):
    region: PuppetRegion = PuppetRegion.US
    device_type: Optional[PuppetDeviceType] = None
    viewport: Optional[Dict] = None

    class Config:
        json_encoders = {
            PuppetRegion: lambda v: v.value,
            PuppetDeviceType: lambda v: v.value
        }

class JobResult(BaseModel):
    id: str
    output: Dict[str, JobTaskOutput]
    start_time: datetime
    stop_time: datetime
    duration: str

class Job(BaseModel):
    id: Optional[str] = None
    url: str
    tasks: List[JobTask]
    schedule: JobSchedule
    navigation: JobNavigation
    puppet_config: Optional[WebAgentConfig] = None

class JobDetail(BaseModel):
    id: str
    name: str
    url: str
    tasks: List[JobTask]
    schedule: JobSchedule
    navigation: JobNavigation
    puppet_config: Optional[WebAgentConfig] = None
    status: JobStatus
    results: List[JobResult] = []
    created_at: datetime

    class Config:
        json_encoders = {JobStatus: lambda v: v.value}

class JobStats(BaseModel):
    total_jobs: int
    completed_jobs: int 
    running_jobs: int
    failed_jobs: int
    token_balance: int


# Re-export all types from individual files
from .enums import (
    JobStatus,
    JobExecutionStatus,
    JobTaskType,
    SpecificDataExtractionType,
    StructuredDataExtractionAttribute,
    SpecificDataExtractionOutputFormat,
    WebScreenCaptureFormat,
    PdfPaperFormat,
    PdfOrientation,
    JobSchedulePolicy,
    JobNavigationType,
    PuppetDeviceType,
    PuppetRegion,
    WebDataSelectorType
)

from .job_schedule_types import (
    JobSchedule
)

from .job_navigation_types import (
    JobNavigation
)

from .job_types import (
    Job,
    JobDetail,
    JobResult,
    JobStats,
    WebAgentConfig
)

from .job_task_types import (
    JobTask,
    JobTaskOutput,
    SpecificDataExtractionOptions,
    WebScreenCaptureOptions,
    WebpageAsMarkdownOptions,
    StructuredDataExtractionOptions,
    StructuredDataExtractionParameter,
    FormDataExtractionOptions,
    TableDataExtractionOptions,
    PresetDataExtractionOptions,
    CustomDataExtractionOptions,
)

from .client_response_types import (
    BaseResponse,
    GetJobResponse,
    GetJobsResponse,
    GetJobResultResponse,
    JobExecutionResponse,
    JobCreationResponse,
    JobEstimatedCostResponse,
    JobStatsResponse
)

from .knowledge_base_types import (
    KnowledgeBase,
    ContextResult,
    CreateKnowledgeBaseResponse,
    GetKnowledgeBasesResponse,
    GetKnowledgeBaseResponse,
    StoreContextResponse,
    FetchContextResponse,
    UpdateContextResponse,
    GatherContextResponse
)

# Optional: Define __all__ to explicitly list what should be exported
__all__ = [
    "JobStatus",
    "JobExecutionStatus",
    "JobTaskType",
    "SpecificDataExtractionType",
    "StructuredDataExtractionAttribute",
    "SpecificDataExtractionOutputFormat",
    "WebScreenCaptureFormat",
    "PdfPaperFormat",
    "PdfOrientation",
    "JobSchedulePolicy",
    "JobNavigationType",
    "PuppetDeviceType",
    "PuppetRegion",
    "WebDataSelectorType",
    "Job",
    "JobSchedule",
    "JobNavigation",
    "JobDetail",
    "JobResult",
    "JobStats",
    "JobTask",
    "JobTaskOutput",
    "SpecificDataExtractionOptions",
    "WebScreenCaptureOptions",
    "WebpageAsMarkdownOptions",
    "StructuredDataExtractionOptions",
    "StructuredDataExtractionParameter",
    "FormDataExtractionOptions",
    "TableDataExtractionOptions",
    "PresetDataExtractionOptions",
    "CustomDataExtractionOptions",
    "BaseResponse",
    "GetJobResponse",
    "GetJobsResponse",
    "GetJobResultResponse",
    "JobCreationResponse",
    "JobExecutionResponse",
    "JobEstimatedCostResponse",
    "JobStatsResponse",
    "WebAgentConfig",
    "KnowledgeBase",
    "ContextResult",
    "CreateKnowledgeBaseResponse",
    "GetKnowledgeBasesResponse",
    "GetKnowledgeBaseResponse",
    "StoreContextResponse",
    "FetchContextResponse",
    "UpdateContextResponse",
    "GatherContextResponse"
]


from enum import Enum

# Enums for Job Status
class JobStatus(str, Enum):
    IDLE = "idle"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    CANCELLED = "cancelled"
    FAILED_TO_SCHEDULE = "failedToSchedule"
    FAILED = "failed"

# Enums for Job Execution Status
class JobExecutionStatus(str, Enum):
    SUCCESS = "success"
    ERROR = "error"

# Enums for Job Task Types
class JobTaskType(str, Enum):
    WEB_PAGE_AS_HTML = "webPageAsHTML"
    WEB_PAGE_METADATA = "webPageMetadata"
    WEB_SCREEN_CAPTURE = "webScreenCapture"
    WEB_PAGE_AS_MARKDOWN = "webPageAsMarkdown"
    SPECIFIC_DATA_EXTRACTION = "specificDataExtraction"

# Enums for Specific Data Extraction Types
class SpecificDataExtractionType(str, Enum):
    STRUCTURED_DATA = "structuredData"
    FORM_DATA = "formData"
    TABLE_DATA = "tableData"
    URLS = "urls"
    IMAGE_LINKS = "imageLinks"
    VIDEO_LINKS = "videoLinks"
    FILE_LINKS = "fileLinks"
    HEADINGS = "headings"
    PHONE_NUMBERS = "phoneNumbers"
    EMAIL_ADDRESSES = "emailAddresses"
    SOCIAL_MEDIA_LINKS = "socialMediaLinks"
    CUSTOM = "custom"

# Enums for Structured Data Extraction Attributes
class StructuredDataExtractionAttribute(str, Enum):
    INNER_TEXT = "innerText"
    TEXT_CONTENT = "textContent"
    HREF = "href"
    SRC = "src"
    ALT = "alt"
    TITLE = "title"
    VALUE = "value"
    CLASS = "class"
    ID = "id"
    NAME = "name"
    TYPE = "type"
    PLACEHOLDER = "placeholder"
    DATA = "data"
    ARIA_LABEL = "ariaLabel"
    TARGET = "target"
    REL = "rel"
    CHECKED = "checked"
    SELECTED = "selected"
    CONTENT = "content"
    PROPERTY = "property"
    CUSTOM = "custom"

# Enums for Specific Data Extraction Output Formats
class SpecificDataExtractionOutputFormat(str, Enum):
    SINGLE_LIST = "singleList"
    SPLIT_BY_PAGE = "splitByPage"

# Enums for Web Screen Capture Formats
class WebScreenCaptureFormat(str, Enum):
    WEBP = "WEBP"
    JPEG = "JPEG"
    PNG = "PNG"
    PDF = "PDF"

# Enums for PDF Paper Formats
class PdfPaperFormat(str, Enum):
    LETTER = "letter"
    LEGAL = "legal"
    TABLOID = "tabloid"
    LEDGER = "ledger"
    A0 = "a0"
    A1 = "a1"
    A2 = "a2"
    A3 = "a3"
    A4 = "a4"
    A5 = "a5"
    A6 = "a6"

# Enums for PDF Orientation
class PdfOrientation(str, Enum):
    LANDSCAPE = "Landscape"
    PORTRAIT = "Portrait"

# Enums for Job Schedule Policies
class JobSchedulePolicy(str, Enum):
    REMAIN_IDLE = "remainIdle"
    IMMEDIATELY = "immediately"
    ONCE_AT = "onceAt"

# Enums for Job Navigation Types
class JobNavigationType(str, Enum):
    NONE = "none"
    LINEAR = "linear"
    BREADTH_FIRST = "breadthFirst"
    CUSTOM_PATH = "customPath"

# Enums for Puppet Device Types
class PuppetDeviceType(str, Enum):
    GENERIC = "Generic"
    BLACKBERRY_PLAYBOOK = "Blackberry PlayBook"
    BLACKBERRY_PLAYBOOK_LANDSCAPE = "Blackberry PlayBook landscape"
    BLACKBERRY_Z30 = "BlackBerry Z30"
    BLACKBERRY_Z30_LANDSCAPE = "BlackBerry Z30 landscape"
    GALAXY_NOTE_3 = "Galaxy Note 3"
    GALAXY_NOTE_3_LANDSCAPE = "Galaxy Note 3 landscape"
    GALAXY_NOTE_II = "Galaxy Note II"
    GALAXY_NOTE_II_LANDSCAPE = "Galaxy Note II landscape"
    GALAXY_S_III = "Galaxy S III"
    GALAXY_S_III_LANDSCAPE = "Galaxy S III landscape"
    GALAXY_S5 = "Galaxy S5"
    GALAXY_S5_LANDSCAPE = "Galaxy S5 landscape"
    GALAXY_S8 = "Galaxy S8"
    GALAXY_S8_LANDSCAPE = "Galaxy S8 landscape"
    GALAXY_S9_PLUS = "Galaxy S9+"
    GALAXY_S9_PLUS_LANDSCAPE = "Galaxy S9+ landscape"
    GALAXY_TAB_S4 = "Galaxy Tab S4"
    GALAXY_TAB_S4_LANDSCAPE = "Galaxy Tab S4 landscape"
    IPAD = "iPad"
    IPAD_LANDSCAPE = "iPad landscape"
    IPAD_GEN_6 = "iPad (gen 6)"
    IPAD_GEN_6_LANDSCAPE = "iPad (gen 6) landscape"
    IPAD_GEN_7 = "iPad (gen 7)"
    IPAD_GEN_7_LANDSCAPE = "iPad (gen 7) landscape"
    IPAD_MINI = "iPad Mini"
    IPAD_MINI_LANDSCAPE = "iPad Mini landscape"
    IPAD_PRO = "iPad Pro"
    IPAD_PRO_LANDSCAPE = "iPad Pro landscape"
    IPAD_PRO_11 = "iPad Pro 11"
    IPAD_PRO_11_LANDSCAPE = "iPad Pro 11 landscape"
    IPHONE_4 = "iPhone 4"
    IPHONE_4_LANDSCAPE = "iPhone 4 landscape"
    IPHONE_5 = "iPhone 5"
    IPHONE_5_LANDSCAPE = "iPhone 5 landscape"
    IPHONE_6 = "iPhone 6"
    IPHONE_6_LANDSCAPE = "iPhone 6 landscape"
    IPHONE_6_PLUS = "iPhone 6 Plus"
    IPHONE_6_PLUS_LANDSCAPE = "iPhone 6 Plus landscape"
    IPHONE_7 = "iPhone 7"
    IPHONE_7_LANDSCAPE = "iPhone 7 landscape"
    IPHONE_7_PLUS = "iPhone 7 Plus"
    IPHONE_7_PLUS_LANDSCAPE = "iPhone 7 Plus landscape"
    IPHONE_8 = "iPhone 8"
    IPHONE_8_LANDSCAPE = "iPhone 8 landscape"
    IPHONE_8_PLUS = "iPhone 8 Plus"
    IPHONE_8_PLUS_LANDSCAPE = "iPhone 8 Plus landscape"
    IPHONE_SE = "iPhone SE"
    IPHONE_SE_LANDSCAPE = "iPhone SE landscape"
    IPHONE_X = "iPhone X"
    IPHONE_X_LANDSCAPE = "iPhone X landscape"
    IPHONE_XR = "iPhone XR"
    IPHONE_XR_LANDSCAPE = "iPhone XR landscape"
    IPHONE_11 = "iPhone 11"
    IPHONE_11_LANDSCAPE = "iPhone 11 landscape"
    IPHONE_11_PRO = "iPhone 11 Pro"
    IPHONE_11_PRO_LANDSCAPE = "iPhone 11 Pro landscape"
    IPHONE_11_PRO_MAX = "iPhone 11 Pro Max"
    IPHONE_11_PRO_MAX_LANDSCAPE = "iPhone 11 Pro Max landscape"
    IPHONE_12 = "iPhone 12"
    IPHONE_12_LANDSCAPE = "iPhone 12 landscape"
    IPHONE_12_PRO = "iPhone 12 Pro"
    IPHONE_12_PRO_LANDSCAPE = "iPhone 12 Pro landscape"
    IPHONE_12_PRO_MAX = "iPhone 12 Pro Max"
    IPHONE_12_PRO_MAX_LANDSCAPE = "iPhone 12 Pro Max landscape"
    IPHONE_12_MINI = "iPhone 12 Mini"
    IPHONE_12_MINI_LANDSCAPE = "iPhone 12 Mini landscape"
    IPHONE_13 = "iPhone 13"
    IPHONE_13_LANDSCAPE = "iPhone 13 landscape"
    IPHONE_13_PRO = "iPhone 13 Pro"
    IPHONE_13_PRO_LANDSCAPE = "iPhone 13 Pro landscape"
    IPHONE_13_PRO_MAX = "iPhone 13 Pro Max"
    IPHONE_13_PRO_MAX_LANDSCAPE = "iPhone 13 Pro Max landscape"
    IPHONE_13_MINI = "iPhone 13 Mini"
    IPHONE_13_MINI_LANDSCAPE = "iPhone 13 Mini landscape"
    IPHONE_14 = "iPhone 14"
    IPHONE_14_LANDSCAPE = "iPhone 14 landscape"
    IPHONE_14_PLUS = "iPhone 14 Plus"
    IPHONE_14_PLUS_LANDSCAPE = "iPhone 14 Plus landscape"
    IPHONE_14_PRO = "iPhone 14 Pro"
    IPHONE_14_PRO_LANDSCAPE = "iPhone 14 Pro landscape"
    IPHONE_14_PRO_MAX = "iPhone 14 Pro Max"
    IPHONE_14_PRO_MAX_LANDSCAPE = "iPhone 14 Pro Max landscape"
    IPHONE_15 = "iPhone 15"
    IPHONE_15_LANDSCAPE = "iPhone 15 landscape"
    IPHONE_15_PLUS = "iPhone 15 Plus"
    IPHONE_15_PLUS_LANDSCAPE = "iPhone 15 Plus landscape"
    IPHONE_15_PRO = "iPhone 15 Pro"
    IPHONE_15_PRO_LANDSCAPE = "iPhone 15 Pro landscape"
    IPHONE_15_PRO_MAX = "iPhone 15 Pro Max"
    IPHONE_15_PRO_MAX_LANDSCAPE = "iPhone 15 Pro Max landscape"
    JIOPHONE_2 = "JioPhone 2"
    JIOPHONE_2_LANDSCAPE = "JioPhone 2 landscape"
    KINDLE_FIRE_HDX = "Kindle Fire HDX"
    KINDLE_FIRE_HDX_LANDSCAPE = "Kindle Fire HDX landscape"
    LG_OPTIMUS_L70 = "LG Optimus L70"
    LG_OPTIMUS_L70_LANDSCAPE = "LG Optimus L70 landscape"
    MICROSOFT_LUMIA_550 = "Microsoft Lumia 550"
    MICROSOFT_LUMIA_950 = "Microsoft Lumia 950"
    MICROSOFT_LUMIA_950_LANDSCAPE = "Microsoft Lumia 950 landscape"
    NEXUS_10 = "Nexus 10"
    NEXUS_10_LANDSCAPE = "Nexus 10 landscape"
    NEXUS_4 = "Nexus 4"
    NEXUS_4_LANDSCAPE = "Nexus 4 landscape"
    NEXUS_5 = "Nexus 5"
    NEXUS_5_LANDSCAPE = "Nexus 5 landscape"
    NEXUS_5X = "Nexus 5X"
    NEXUS_5X_LANDSCAPE = "Nexus 5X landscape"
    NEXUS_6 = "Nexus 6"
    NEXUS_6_LANDSCAPE = "Nexus 6 landscape"
    NEXUS_6P = "Nexus 6P"
    NEXUS_6P_LANDSCAPE = "Nexus 6P landscape"
    NEXUS_7 = "Nexus 7"
    NEXUS_7_LANDSCAPE = "Nexus 7 landscape"
    NOKIA_LUMIA_520 = "Nokia Lumia 520"
    NOKIA_LUMIA_520_LANDSCAPE = "Nokia Lumia 520 landscape"
    NOKIA_N9 = "Nokia N9"
    NOKIA_N9_LANDSCAPE = "Nokia N9 landscape"
    PIXEL_2 = "Pixel 2"
    PIXEL_2_LANDSCAPE = "Pixel 2 landscape"
    PIXEL_2_XL = "Pixel 2 XL"
    PIXEL_2_XL_LANDSCAPE = "Pixel 2 XL landscape"
    PIXEL_3 = "Pixel 3"
    PIXEL_3_LANDSCAPE = "Pixel 3 landscape"
    PIXEL_4 = "Pixel 4"
    PIXEL_4_LANDSCAPE = "Pixel 4 landscape"
    PIXEL_4A_5G = "Pixel 4a (5G)"
    PIXEL_4A_5G_LANDSCAPE = "Pixel 4a (5G) landscape"
    PIXEL_5 = "Pixel 5"
    PIXEL_5_LANDSCAPE = "Pixel 5 landscape"
    MOTO_G4 = "Moto G4"
    MOTO_G4_LANDSCAPE = "Moto G4 landscape"

# Enums for Puppet Regions
class PuppetRegion(str, Enum):
    US = "US"
    DE = "DE"
    FR = "FR"
    IT = "IT"
    UA = "UA"

# Enums for Web Data Selector Types
class WebDataSelectorType(str, Enum):
    NONE = "none"
    CSS = "css"
    XPATH = "xpath"



import requests;
from typing import Dict, List;
from ..types.enums import (
    PuppetRegion, 
);
from ..types import (
    WebAgentConfig, 
    Job, JobSchedule, JobNavigation,
    JobNavigationType, JobSchedulePolicy, 
    BaseResponse, JobExecutionResponse, JobExecutionStatus,
    GetJobResponse, GetJobsResponse, GetJobResultResponse,
    JobEstimatedCostResponse, JobStatsResponse, JobCreationResponse,
    CreateKnowledgeBaseResponse, GetKnowledgeBasesResponse, GetKnowledgeBaseResponse,
    StoreContextResponse, FetchContextResponse, UpdateContextResponse, GatherContextResponse
);
from .utils import dict_to_camel_case, dict_to_snake_case

# Grabba SDK Class
class Grabba:
    def __init__(self, api_key: str, region: PuppetRegion = PuppetRegion.US.value):
        if not api_key:
            raise ValueError("API key is required")
        self.api_key = api_key
        self.api_url = "https://api.grabba.dev/v1"
        self.default_puppet_config = WebAgentConfig(region=region)
        self.default_job_navigation = JobNavigation(type=JobNavigationType.NONE.value)
        self.default_job_schedule = JobSchedule(policy=JobSchedulePolicy.IMMEDIATELY.value)

    def _get_headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
            "X-Api-Key": self.api_key,
        }
    
    def get_stats(self) -> JobStatsResponse:
        response = requests.get(
            f"{self.api_url}/stats",
            headers=self._get_headers(),
        )
        # Validate response
        response.raise_for_status()
        response_obj = response.json()
        return JobStatsResponse(
            message=response_obj["message"],
            job_stats=dict_to_snake_case(response_obj["jobStats"])
        )

    def estimate_job_cost(self, job: Job) -> JobEstimatedCostResponse:
        job = Job.model_validate(job)
        if not job.puppet_config:
            job.puppet_config = self.default_puppet_config
        if not job.navigation:
            job.navigation = self.default_job_navigation
        if not job.schedule:
            job.schedule = self.default_job_schedule
        job_data = job.model_dump(by_alias=True, exclude_none=True)
        # convert job object keys to camelCase
        jobToCamelCase = dict_to_camel_case(job_data)
        # Send request
        response = requests.post(
            f"{self.api_url}/estimate-job-cost",
            headers=self._get_headers(),
            json=jobToCamelCase,
        )
        # Return response for only 'BadRequest' and 'response.OK'
        if response.ok:
            response_obj = response.json()
            return JobEstimatedCostResponse(
                message=response_obj['message'], 
                job_estimated_cost=response_obj['jobEstimatedCost']
            )
        if response.status_code == 400:
            # Catch error message
            response_obj = response.json()
            job_errors = { "errors": response_obj }
            return JobExecutionResponse(
                message="BadRequestError", 
                job_estimated_cost=job_errors
            )
        return response.raise_for_status()

    def create_job(self, job: Job) -> JobExecutionResponse:
        job = Job.model_validate(job)
        # Ensure job is only created, not executed.
        job.schedule = JobSchedule(policy=JobSchedulePolicy.REMAIN_IDLE.value)
        if not job.puppet_config:
            job.puppet_config = self.default_puppet_config
        if not job.navigation:
            job.navigation = self.default_job_navigation
        job_data = job.model_dump(by_alias=True, exclude_none=True)
        # convert job object keys to camelCase
        jobToCamelCase = dict_to_camel_case(job_data)
        # Send request
        response = requests.post(
            f"{self.api_url}/create-job",
            headers=self._get_headers(),
            json=jobToCamelCase,
        )
        # Return response for only 'BadRequest' and 'response.OK'
        if response.ok:
            response_obj = response.json()
            return JobCreationResponse(
                message=response_obj['message'],
                job=dict_to_snake_case(response_obj['job'])
            )
        if response.status_code == 400:
            # Catch error message
            response_obj = response.json()
            job_errors = { "errors": response_obj }
            return JobCreationResponse(
                message="BadRequestError", 
                job=job_errors
            )
        return response.raise_for_status()
    
    def extract(self, job: Job) -> JobExecutionResponse:
        job = Job.model_validate(job)
        if not job.puppet_config:
            job.puppet_config = self.default_puppet_config
        if not job.navigation:
            job.navigation = self.default_job_navigation
        if not job.schedule:
            job.schedule = self.default_job_schedule
        job_data = job.model_dump(by_alias=True, exclude_none=True)
        # convert job object keys to camelCase
        jobToCamelCase = dict_to_camel_case(job_data)
        # Send request
        response = requests.post(
            f"{self.api_url}/extract",
            headers=self._get_headers(),
            json=jobToCamelCase,
        )
        # Return response for only 'BadRequest' and 'response.OK'
        if response.ok:
            response_obj = response.json()
            return JobExecutionResponse(
                status=response_obj['status'], 
                message=response_obj['message'], 
                job_result=dict_to_snake_case(
                    response_obj['jobResult'],
                    skip_regex=r'^\d+\s-\s.+$' # Skips task output key
                )
            )
        if response.status_code == 400:
            # Catch error message
            response_obj = response.json()
            job_errors = { "errors": response_obj }
            return JobExecutionResponse(
                status=JobExecutionStatus.ERROR,
                message="BadRequestError", 
                job_result=job_errors
            )
        return response.raise_for_status()
    
    def schedule_job(self, job_id: str) -> JobExecutionResponse:
        response = requests.post(
            f"{self.api_url}/schedule-job/{job_id}",
            headers=self._get_headers(),
        )
        # Return response for only 'BadRequest' and 'response.OK'
        if response.ok:
            response_obj = response.json()
            return JobExecutionResponse(
                status=response_obj['status'], 
                message=response_obj['message'], 
                job_result=dict_to_snake_case(
                    response_obj['jobResult'],
                    skip_regex=r'^\d+\s-\s.+$' # Skips task output key
                )
            )
        if response.status_code == 400:
            # Catch error message
            response_obj = response.json()
            job_errors = { "errors": response_obj }
            return JobExecutionResponse(
                status=JobExecutionStatus.ERROR,
                message="BadRequestError", 
                job_result=job_errors
            )
        return response.raise_for_status()

    def get_jobs(self, page=1, limit=25) -> GetJobsResponse:
        response = requests.get(
            f"{self.api_url}/jobs?page={page}&limit={limit}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        response_obj = response.json()
        return GetJobsResponse(
            message=response_obj['message'],
            jobs=dict_to_snake_case(
                response_obj['jobs'],
                skip_regex=r'^\d+\s-\s.+$'
            ),
            pagination=dict_to_snake_case(
                response_obj['pagination']
            )
        )

    def get_job(self, job_id: str) -> GetJobResponse:
        response = requests.get(
            f"{self.api_url}/jobs/{job_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        response_obj = response.json()
        return GetJobResponse(
            message=response_obj["message"],
            job=dict_to_snake_case(response_obj['job'])
        )
    
    def delete_job(self, job_id: str) -> BaseResponse:
        response = requests.delete(
            f"{self.api_url}/jobs/{job_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return BaseResponse(**response.json())

    def get_job_result(self, job_result_id: str) -> GetJobResultResponse:
        response = requests.get(
            f"{self.api_url}/job-result/{job_result_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        response_obj = response.json()
        return GetJobResultResponse(
                message=response_obj['message'], 
                job_result=dict_to_snake_case(
                    response_obj['jobResult'],
                    skip_regex=r'^\d+\s-\s.+$' # Skips task output key
                )
            )
    
    def delete_job_result(self, job_result_id: str) -> BaseResponse:
        response = requests.delete(
            f"{self.api_url}/job-result/{job_result_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return BaseResponse(**response.json())

    def get_available_regions(self) -> List[Dict[str, PuppetRegion]]:
        return [{k: v.value} for k, v in PuppetRegion.__members__.items()]

    # Knowledge Base Methods
    def create_knowledge_base(self, name: str, description: str = None) -> CreateKnowledgeBaseResponse:
        payload = {"name": name}
        if description:
            payload["description"] = description
        
        response = requests.post(
            f"{self.api_url}/knowledge-bases",
            headers=self._get_headers(),
            json=payload,
        )
        response.raise_for_status()
        return CreateKnowledgeBaseResponse(**response.json())

    def get_knowledge_bases(self) -> GetKnowledgeBasesResponse:
        response = requests.get(
            f"{self.api_url}/knowledge-bases",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return GetKnowledgeBasesResponse(**response.json())

    def get_knowledge_base(self, kb_id: str) -> GetKnowledgeBaseResponse:
        response = requests.get(
            f"{self.api_url}/knowledge-bases/{kb_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return GetKnowledgeBaseResponse(**response.json())

    def delete_knowledge_base(self, kb_id: str) -> BaseResponse:
        response = requests.delete(
            f"{self.api_url}/knowledge-bases/{kb_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return BaseResponse(**response.json())

    def store_context(self, kb_id: str, context: str, metadata: Dict = None) -> StoreContextResponse:
        payload = {"context": context}
        if metadata:
            payload["metadata"] = metadata

        response = requests.post(
            f"{self.api_url}/knowledge-bases/{kb_id}/context",
            headers=self._get_headers(),
            json=payload,
        )
        response.raise_for_status()
        return StoreContextResponse(**response.json())

    def fetch_context(self, kb_id: str, query: str, options: Dict = None) -> FetchContextResponse:
        payload = {"query": query}
        if options:
            payload["options"] = options

        response = requests.get(
            f"{self.api_url}/knowledge-bases/{kb_id}/context",
            headers=self._get_headers(),
            json=payload,
        )
        response.raise_for_status()
        return FetchContextResponse(**response.json())

    def update_context(self, context_id: str, content: str = None, metadata: Dict = None) -> UpdateContextResponse:
        payload = {}
        if content:
            payload["content"] = content
        if metadata:
            payload["metadata"] = metadata

        response = requests.put(
            f"{self.api_url}/knowledge-bases/context/{context_id}",
            headers=self._get_headers(),
            json=payload,
        )
        response.raise_for_status()
        return UpdateContextResponse(**response.json())

    def delete_context(self, context_id: str) -> BaseResponse:
        response = requests.delete(
            f"{self.api_url}/knowledge-bases/context/{context_id}",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return BaseResponse(**response.json())

    def gather_context(self, kb_id: str) -> GatherContextResponse:
        response = requests.get(
            f"{self.api_url}/knowledge-bases/{kb_id}/gather-context",
            headers=self._get_headers(),
        )
        response.raise_for_status()
        return GatherContextResponse(**response.json())
    



import re

def camel_to_snake(name: str) -> str:
    """Convert camelCase string to snake_case"""
    # Handle acronyms (e.g., 'JSONData' → 'json_data')
    name = re.sub('([A-Z]+)([A-Z][a-z])', r'\1_\2', name)
    # Handle normal camelCase (e.g., 'camelCase' → 'camel_case')
    name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', name)
    return name.lower()

def to_camel_case(snake_str):
    """Convert snake_case to camelCase"""
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])

def dict_to_camel_case(data, skip_regex=None):
    """
    Recursively convert dictionary keys to camelCase.
    skip_regex: Keys matching this pattern won't be converted (but their children will be)
    """
    if isinstance(data, dict):
        new_dict = {}
        for key, value in data.items():
            current_key_skipped = bool(skip_regex and re.match(skip_regex, str(key)))
            # Only skip conversion of THIS key if it matches the pattern
            new_key = key if current_key_skipped else to_camel_case(key)
            # Run "dict_to_camel_case" recursively
            new_dict[new_key] = dict_to_camel_case(
                value, 
                skip_regex=skip_regex,
            )
        return new_dict
    elif isinstance(data, list):
        return [dict_to_camel_case(item, skip_regex) for item in data]
    else:
        return data
    
def dict_to_snake_case(data, skip_regex=None):
    """
    Recursively convert dictionary keys to snake_case.
    skip_regex: Keys matching this pattern won't be converted (but their children will be)
    parent_key_skipped: Internal flag to track if we're under a skipped key
    """
    if isinstance(data, dict):
        new_dict = {}
        for key, value in data.items():
            current_key_skipped = bool(skip_regex and re.match(skip_regex, str(key)))
            # Only skip conversion of THIS key if it matches the pattern
            new_key = key if current_key_skipped else camel_to_snake(key)            
            # Run "dict_to_snake_case" recursively
            new_dict[new_key] = dict_to_snake_case(
                value, 
                skip_regex=skip_regex
            )
        return new_dict
    elif isinstance(data, list):
        return [dict_to_snake_case(item, skip_regex) for item in data]
    else:
        return data



from typing import List, Dict, Optional
from pydantic import BaseModel, Field

class KnowledgeBase(BaseModel):
    id: str = Field(..., alias='_id')
    user_id: str = Field(..., alias='userId')
    name: str
    description: Optional[str] = None
    size_in_bytes: int = Field(..., alias='sizeInBytes')
    max_size_in_bytes: int = Field(..., alias='maxSizeInBytes')
    created_at: str = Field(..., alias='createdAt')

class ContextResult(BaseModel):
    id: str = Field(..., alias='_id')
    kb_id: str = Field(..., alias='kbId')
    source: str
    source_type: str = Field(..., alias='sourceType')
    chunk_index: int = Field(..., alias='chunkIndex')
    content: str
    metadata: Dict
    created_at: str = Field(..., alias='createdAt')

class CreateKnowledgeBaseResponse(BaseModel):
    message: str
    knowledge_base: KnowledgeBase = Field(..., alias='knowledgeBase')

class GetKnowledgeBasesResponse(BaseModel):
    message: str
    knowledge_bases: List[KnowledgeBase] = Field(..., alias='knowledgeBases')

class GetKnowledgeBaseResponse(BaseModel):
    message: str
    knowledge_base: KnowledgeBase = Field(..., alias='knowledgeBase')

class StoreContextResponse(BaseModel):
    message: str
    context_ids: List[str] = Field(..., alias='contextIds')

class FetchContextResponse(BaseModel):
    message: str
    results: List[ContextResult]

class UpdateContextResponse(BaseModel):
    message: str
    context: ContextResult

class GatherContextResponse(BaseModel):
    message: str
    metadata: Dict
